/* seqidx.c ----------- utilities for accessing the contig sequence data thorugh
   the seqindex file */

#include "psublast.h"
#include <sys/types.h>
#include <sys/stat.h>
#include "seqidx.h"  

/* externs */
extern char der_datadir[];
extern char bas_datadir[];

/* interface */
SEQ *get_sequence(char *bug, char *alias, char *contig_name, int *idx,int flag);
SEQ *get_indexed_seq(char *bug,char *alias,char *contig_name,int *idx,int flag);
SEQ *get_sequential_seq(char *bug, char *alias, char *contig_name);

/* in this file */
static int  seqidx_get_entry(char *,char *,char *,int *,long int *,int);
static int  seqidx_bin_search(FILE *,long int,int,long int *,char *);
static int  seqidx_parse_header(char *,char *,FILE *,char **,long int *);
static int  seqidx_parse_line(char *line, int *idx, long int *fpos, char **key);
static void seqidx_find_midrow(FILE *fp,long int *first,long int *last,int *Ix,int ix);



/* returns a SEQ* object refering to the combination (contig_name,idx);
   if flag==IN_FLAG, locates the contig by name in the index file (if
   valid), and returns the index number in idx. If flag==IN_FLAG, attempts
   to locate the contig by idx first. If access through index fails, performs
   sequential search in the sequence file.
*/      

SEQ *get_sequence(char *bug, char *alias, char *contig_name, int *idx, int flag)
{
   SEQ *sf;
        
   if ((sf=get_indexed_seq(bug, alias, contig_name, idx, flag))==NULL)
        return (sf=get_sequential_seq(bug, alias, contig_name));
        
   return sf;
}       


/* returns a SEQ* object, by sequentially searching the sequence data file for
   the given contig name; returns NULL in case of failure  */

SEQ *get_sequential_seq(char *bug, char *alias, char *contig_name)
{
   char buf[500], *key, *s, *q, contig_key[200], *h;
   SEQ *sf;

   s = contig_name; q = contig_key;
   while (*s && !isspace(*s)) *q++ = *s++;
   *q = '\0';

   /* read sequentially from sequence file */

   sprintf(buf, "%s/%s", bas_datadir, bug);
   sf = seq_open(buf, 0, 0);

   while (seq_read(sf)>0) {
      h=SEQ_HEAD(sf);
      key = strtok(h, " \n\t");
      if (!strcmp(key, contig_key))
           return sf;
      free(h);
   }
   seq_close(sf);
   return NULL;
}


/* returns a SEQ* object refering to the combination (contig_name,idx),
   using the seqindex file; returns NULL in case of failure              
*/
SEQ *get_indexed_seq(char *bug, char *alias, char *contig_name, int *idx, int flag)
{       
   long int fpos;
   char buf[500];
   SEQ *sf;

   /* find if the index file is valid and extract contig index num */

   if ((seqidx_get_entry(bug, alias, contig_name, idx, &fpos, flag)<=0) || (*idx<=0))
       return NULL;

   sprintf(buf, "%s/%s", bas_datadir, bug);
   sf = seq_open(buf, 0, 0);
   seq_revisit(sf, fpos);
   if (seq_read(sf)<=0) {
       fprintf(stderr, "Inconsistent sequence and seqindex files for %s (%s).",
                       contig_name, alias);
       return NULL;
   }
   return sf;
} 


/* attempts to locate a contig by index num/name in the seqindex file; returns 
   -1 if the index file is non-existant or is invalid, 1 in case of success
*/

static int seqidx_get_entry(char *bug, char *alias, char *contig_name, int *idx,
                         long int *fpos, int flag)
{  
   char buf[500], *s, *q, *key, *contig_prefix=NULL, contig_key[200];
   FILE *fp;
   long int last, tmpi;
   
   *idx = -1;
   s = contig_name; q = contig_key;
   while (*s && !isspace(*s)) *q++ = *s++;
   *q = '\0';  
   
   sprintf(buf, "%s/%s.seqindex", der_datadir, alias);
      
   if ((fp=fopen(buf, "r"))==NULL) {
        /* no index file */
       fprintf(stderr, "Index file not found for %s.\n", alias);
       return (*idx=-1);
   }

   if (seqidx_parse_header(bug, alias, fp, &contig_prefix, &last)<=0) {
       fprintf(stderr, "Seqindex file invalid for %s.\n", alias);
       fclose(fp);
       return (*idx=-1);
   }
   
   /* read header successfully */
   if (contig_prefix != NULL) {
       /* check if contig name conforms to the prefix rule */
       if (strncmp(contig_key,contig_prefix,strlen(contig_prefix))) {
           tmpi = atoi(contig_key+strlen(contig_prefix));
           if ((flag!=IN_FLAG) || (tmpi!=*idx) || (*idx<=0))
             *idx = -1;
       } else { 
            /* prefix rule doesn't work, search by index or contig name */
            *idx = -1; 
       }
   } else {
       /* no prefix, but idx may be either knwon or unknown */
       if  ((flag!=IN_FLAG) || (tmpi!=*idx) || (*idx>0)) *idx = -1;
   }
   
   /* retrieve actual entry: use binary search if idx>0, sequential
      search otherwise */
   if (*idx>0) {
       if (seqidx_bin_search(fp, last, *idx, fpos, contig_prefix)<=0)
           fclose(fp); return (*idx = -1);
   } else {
      *idx = *fpos = -1;
      while (fgets(buf, 200, fp)!=NULL) {
         if (seqidx_parse_line(buf, idx, fpos, &key)<=0) {
            fprintf(stderr, "Invalid seqindex file %s.%s\n", alias, buf);
            fclose(fp); return -1;
         }   
         if (!strcmp(key, contig_key)) {
             fclose(fp); return *idx;
         }  
     }  
   } 
   fclose(fp);
   return -1;
}  


/* checks the header of the seqindex file; returns -1 if invalid, 1 otherwise;
   also retrieves the automatic contig prefix, if any, for faster access */

static int seqidx_parse_header(char *bug, char *alias, FILE *fp, char **contig_prefix, long int *last)
{  
   char buf[1000], *s, *q;
   off_t stsize;
   struct stat st;
      
   if ((fgets(buf, 1000, fp)==NULL) || strncmp(buf, "#seqindex:", 10))
               fatal("Not a sequence index file.");
   if ((s = strstr(buf, "prefix"))!=NULL) {
       s += strlen("prefix");
       while (isspace(*s)) s++;
       q = *contig_prefix = (char *)ckalloc(strlen(s)+1);
       while (!isspace(*s)) *q++ = *s++; *q = '\0';
   }     
     
   fgets(buf, 1000, fp);
   if (sscanf(buf, " %ld seqindex", last)!=1) {
       fprintf(stderr, "Warning: Not a seqindex file (%s).\n", alias);
       return -1;
   }
   fgets(buf, 1000, fp);
   sscanf(buf," %ld", &stsize);
   sprintf(buf, "%s/%s", bas_datadir, bug);
   stat(buf, &st); 
   if (st.st_size!=stsize) {
       fprintf(stderr, "Warning: Stale seqindex file (%s).", buf);
       return -1;
   }
               
   return 1;
}      
       
/* parses one line of the seqindex file */
static int seqidx_parse_line(char *line, int *idx, long int *fpos, char **key)
{      
   char *s;
     
   if (sscanf(line, "%d %ld", idx, fpos)!=2) return -1;
   s = line;
   while (isspace(*s)) ++s; /* skip over fisrt spaces, if any */
   while (isdigit(*s)) ++s; /* skip over index num */
   while (isspace(*s)) ++s;
   while (isdigit(*s)) ++s; /* skip over fpos */
   while (isspace(*s)) ++s;
   
   *key = strtok(s, "\n\t");
   
   return 1;
}

/* performs the partial binary search for the seqindex file */
static int seqidx_bin_search(FILE *fp, long int last, int idx,
                                long int *fpos, char *contig_prefix)
{      
   long int first = ftell(fp);
   int Idx = 0;
   char buf[200], tmp[100], *key;
     
   if (last>2000) { 
      seqidx_find_midrow(fp, &first, &last, &Idx, idx);
      if (Idx!=idx) {
          seqidx_find_midrow(fp, &first, &last, &Idx, idx);
          if (Idx!=idx)
              seqidx_find_midrow(fp, &first, &last, &Idx, idx);
      }
   }
   while ((Idx<idx) && (fgets(buf, 200, fp)!=NULL))
      if (sscanf(buf, "%d %ld", &Idx, fpos)!=2) fatal("sscanf failed.");
   while (Idx==idx) {
      int tmpi;
      if (seqidx_parse_line(buf,&tmpi,fpos,&key)<=0) return -1;
      sprintf(tmp, "%s%d", contig_prefix, idx);
      if (!strcmp(tmp, key))
          return 1;             
      if (fgets(buf, 200, fp)==NULL) break;
   }
   
   return -1;
}    

static void seqidx_find_midrow(FILE *fp, long int *first, long int *last, int *Idx, int idx)   
{       
     char c, tmp[50];
     int i=0;
     long int mid;
           
     mid = (int)((*last+*first)/2);
     fseek(fp, mid, SEEK_SET);
     while ((c=getc(fp))!='\n');
     while (isdigit(c=getc(fp))) tmp[i++] = c;
     tmp[i] = '\0';
     ungetc(c, fp);     
     sscanf(tmp, "%d", Idx);
     if (idx>=*Idx) {
         /* next search is in the upper half */
         while(--i>=0) ungetc(tmp[i], fp);
         *first = ftell(fp);
     } else {
         /* next search is in the first half */
         fseek(fp, *first, SEEK_SET);
         *last =  mid;
     }  
}       

